"""Make sure the analysis group works."""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import pytest
from typedargs.exceptions import ArgumentError
from iotile_analytics.core import CloudSession, AnalysisGroup
from iotile_analytics.core.stream_series import StreamSeries
from iotile_analytics.core.exceptions import AuthenticationError, CertificateVerificationError


def test_session_login(water_meter):
    """Make sure we can login successfully."""

    domain, _cloud = water_meter

    session = CloudSession('test@arch-iot.com', 'test', domain=domain, verify=False)
    assert session.token == "JWT_USER"

    with pytest.raises(AuthenticationError):
        CloudSession('test@arch-iot.com', 'test2', domain=domain, verify=False)

def test_token_login(water_meter):
    """Make sure we can login successfully with a token"""

    domain, _cloud = water_meter

    session = CloudSession(token="JWT_USER", domain=domain, verify=False)

    assert session.token == "JWT_USER"

    with pytest.raises(AuthenticationError):
        CloudSession(token="WRONG_TOKEN", domain=domain, verify=False)

def test_ssl_verification(water_meter):
    """Make sure we default to verifying SSL partners."""

    domain, _cloud = water_meter

    with pytest.raises(CertificateVerificationError):
        CloudSession('test@arch-iot.com', 'test', domain=domain, verify=True)


def test_multiple_login(water_meter):
    """Make sure multiple logins work correctly."""

    domain, _cloud = water_meter

    CloudSession('test@arch-iot.com', 'test', domain=domain, verify=False)

    session = CloudSession(domain=domain)
    assert session.token == 'JWT_USER'


def test_device_analysis(water_meter):
    """Make sure we can create an analysis group from a device.

    Also tests to make sure we properly read in stream time series and event
    data.  We support reading stream data from json API dumps or from CSV
    files.
    """

    domain, _cloud = water_meter
    CloudSession('test@arch-iot.com', 'test', domain=domain, verify=False)

    group = AnalysisGroup.FromDevice('d--0000-0000-0000-00d2', domain=domain)

    assert group.stream_counts['s--0000-0077--0000-0000-0000-00d2--5001']['points'] == 11
    assert group.stream_counts['s--0000-0077--0000-0000-0000-00d2--5002']['points'] == 3


def test_stream_download(filter_group):
    """Make sure we can download streams from an analysis group."""

    data = filter_group.fetch_stream('5001')
    assert len(data) == 11

    assert isinstance(data, StreamSeries)
    assert set(data.available_units) == set(['Liters', 'Gallons', 'Cubic Meters', 'Acre Feet'])

    outL = data.convert('Liters')
    outG = data.convert('Gallons')

    assert data.iloc[0][0] == pytest.approx(outL.iloc[0][0])
    assert data.iloc[0][0] / 3.78541 == pytest.approx(outG.iloc[0][0])
    with pytest.raises(ArgumentError):
        data.convert('Test Unit')


def test_invalid_stream(filter_group):
    """Make sure we raise the right error if we can't find a stream."""

    with pytest.raises(ArgumentError):
        filter_group.fetch_stream('5500')


def test_raw_events(filter_group):
    """Make sure we can download raw events."""

    meta = filter_group.fetch_events('5001')
    assert meta.columns[0] == 'event_id'
    assert meta.columns[1] == 'has_raw_data'
    assert meta.columns[2] == 'meta'

    assert meta.iloc[0]['meta'] == 'hello'
    assert meta.iloc[1]['meta'] == 'goodbye'

    raw = filter_group.fetch_raw_events('5001')

    assert len(raw) == 2
    assert raw.iloc[0]['test'] == 1
    assert raw.iloc[1]['goodbye'] == 15.0


def test_channel_info(filter_group):
    """Make sure we can download raw events."""

    # Test AnalyisGroup API
    device = filter_group.source_info
    assert device['slug'] == 'd--0000-0000-0000-00d2'
    assert device['label'] == 'Filtration Flow'
    assert device['org'] == 'test_org'

    props = filter_group.properties
    assert props['CargoDescription'] == 'SO# 83469'
    assert props['Country'] == 'KOREA'



"""A Pandas DataFrame subclass for dealing with TimeSeries data from IOTile.cloud streams."""

import pandas as pd
from typedargs.exceptions import ArgumentError


class StreamSeries(pd.DataFrame):
    """A DataFrame subclass for handle stream timeseries data.

    All arguemnts are passed through to the underlying dataframe
    except for the optional stream and variable arguments that let
    you set associated metadata and units for this stream.
    """

    _metadata = ['_stream', '_vartype']

    def __init__(self, *args, **kwargs):
        self._stream = None
        self._vartype = None

        if 'stream' in kwargs:
            self.set_stream(kwargs['stream'])
            del kwargs['stream']

        if 'vartype' in kwargs:
            self.set_variable(kwargs['vartype'])
            del kwargs['vartype']

        super(StreamSeries, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return StreamSeries

    def set_stream(self, stream):
        """Set the stream metadata for this data series.

        This data is what is returned by the iotile.cloud
        API /api/v1/stream/<slug>/

        It contains the stream slug, some metadata and the list
        of available output units for conversion.

        Args:
            stream (dict): Stream metadata returned from iotile.cloud
        """

        self._stream = stream

    def set_vartype(self, vartype):
        """Set the variable type and units metadata for this data series.

        This data is what is returned by the iotile.cloud API
        /api/v1/vartype/<slug>
        """

        self._vartype = vartype

    @property
    def available_units(self):
        """Get the available output units for this data stream.

        Returns:
            list: A list of unit names that this stream can be
                converted to.
        """

        if self._vartype is not None:
            return [x['unit_full'] for x in self._vartype.get('available_output_units', [])]

        if self._stream is not None:
            return [self._stream['output_unit']['unit_full']]

        return []

    def _mdo_for_unit(self, unit):
        if unit not in self.available_units:
            raise ArgumentError("Unknown units", units=unit)

        if self._vartype is not None:
            units = {x['unit_full']: x for x in self._vartype.get('available_output_units', [])}
        else:
            units = {self._stream['output_unit']['unit_full']: self._stream['output_unit']}

        unit_data = units[unit]
        m = unit_data.get('m', 1)
        d = unit_data.get('d', 1)
        o = unit_data.get('o', 0.0)

        return (m, d, o)

    def convert(self, units):
        """Convert this stream to another set of supported units.

        Args:
            units (str): The desired output units for this stream.

        Returns:
            pd.DataFrame: The converted data stream.
        """

        out = pd.DataFrame(self.values, copy=True, index=self.index)

        m, d, o = self._mdo_for_unit(units)

        out[0] *= float(m)
        out[0] /= float(d)
        out[0] += float(o)

        return out



"""Basic configuration for interactive vs script environments."""
from .exceptions import MissingPackageError


class Environment(object):
    """Configuration class for better interactive computing experience."""

    Console = 0
    Notebook = 1
    Script = 2
    Unattended = 3

    _settings = {}

    def __init__(self):
        self.interactivity = Environment._settings.get('interactivity', Environment.Script)

    @classmethod
    def SetupNotebook(cls):
        """Configure iotile_analytics for a Jupyter Notebook."""

        try:
            import tqdm  # pylint:disable=W0612; We just want to test importability
        except ImportError:
            raise MissingPackageError("Missing required package tqdm", package="tqdm", suggestion="pip install tqdm")

        # If we are using bokeh, make sure we set it up for notebook output
        try:
            from bokeh.io import output_notebook
            output_notebook(hide_banner=True)
        except ImportError:
            pass

        cls._settings['interactivity'] = Environment.Notebook

    @classmethod
    def SetupScript(cls):
        """Configure iotile_analytics to run in an attended script.

        Attended scripts are those that are watched by a user in the
        command line but not run using jupyter-console.  They are run
        as normal python scripts.

        No graphical plotting is allowed since the shell doesn't allow
        it and progress bars are displayed in an ascii format.
        """

        try:
            import tqdm  # pylint:disable=W0612; We just want to test importability
        except ImportError:
            raise MissingPackageError("Missing required package tqdm", package="tqdm", suggestion="pip install tqdm")

        cls._settings['interactivity'] = Environment.Script



"""A collection of useful numerical utility functions."""

from .domain import find_domain, combine_domains
from .envelope import envelope
from .aggregator import TimeseriesSelector

__all__ = ['find_domain', 'combine_domains', 'envelope', 'TimeseriesSelector']



"""A class that lets you compare portions of multiple timeseries datasets."""

from builtins import range, str
from typedargs.exceptions import ArgumentError
from iotile_analytics.core.exceptions import UsageError
import pandas as pd
import numpy as np


class TimeseriesSelector(object):
    """A utility class for aggregating and selecting data from multiple timeseries.

    The selector keeps track of the first and last timestamp that was added and lets
    you restrict data to only a given day, week or month while iterating through every
    such segment of the data.

    This is particularly useful for producing coherent views and reports of data that
    may cover different intersecting but not identical time spans.


    Args:
        timezone (str): The timezone to use to localize all data coming in so that days
            correspond correctly to days in that timezone.  Default: UTC.  This should be
            a timzeone value like 'US/Central'.
    """

    def __init__(self, timezone='UTC'):
        self._timezone = timezone

        self.oldest_point = None
        self.newest_point = None

    def add_data(self, data):
        """Add a new dataset to the aggregator.

        The data is homogenized and the underlying domain of the aggregator is
        extended if necessary.  The data itself is not stored.

        Args:
            data (pd.DataFrame): A dataframe to add to our aggregator.
        """

        # First localize the data into the right timezone
        if self._check_tz_naive(data):
            data = data.tz_localize('UTC', copy=False)

        data = data.tz_convert(self._timezone, copy=False)

        oldest = data.index.min()
        newest = data.index.max()

        if self.oldest_point is None or oldest < self.oldest_point:
            self.oldest_point = oldest
        if self.newest_point is None or newest > self.newest_point:
            self.newest_point = newest

    def _check_tz_naive(self, data):
        return data.index.tzinfo is None

    def _generate_index(self, freq):
        if self.oldest_point is None or self.newest_point is None:
            raise UsageError("You must add at least one dataset to the selector first before trying to generate an index.")

        return pd.PeriodIndex(start=self.oldest_point, end=self.newest_point, freq=freq)

    def months(self):
        """Get all of the unique months spanning our data domain.

        The results are returned in time order with duplicates removed.

        Returns:
            pd.PeriodIndex
        """

        return self._generate_index('M')

    def weeks(self):
        """Get all unique weeks spanning our data domain.

        The results are returned in time order with duplicates removed.

        Returns:
            pd.PeriodIndex
        """

        return self._generate_index('W')

    def days(self):
        """Get all unique days spanning our data domain.

        The results are returned in time order with duplicates removed.

        Returns:
            list of datemasks: The day date masks that can be used to select a subset
                of data from this aggregator.
        """

        return self._generate_index('D')

    def divide_period(self, period, freq):
        """Turn a single period into a PeriodIndex with a given frequency.

        For example you could subdivide a month into days or a week into hours.
        The subdivision will be performed down to second precision by default.

        Args:
            period (pd.Period): The period that we wish to subdivid
            freq (str): A Pandas frequency string to use to determine how big the
                chunks are that we are going to subdivid period into.

        Returns:
            pd.DatetimeIndex: The subdivided period.
        """

        start = period.to_timestamp('S', 'S')
        end = period.to_timestamp('S', 'E')

        return pd.DatetimeIndex(start=start, end=end, freq=freq, tz=self._timezone)

    def resample(self, data, period, freq, resample='last'):
        """Convenience function to restrict and resample a data set.

        The data set is restricted to the indicated period and if necessary
        resampled to appropriate frequency.  If the data is already sampled
        at the correct frequency then no resampling occurs and NaNs are used
        to fill in any gaps in the indicated period.

        Naive timestamps in data are assumed to be in UTC and localized to
        the configured timezone of this selector before resampling.

        Args:
            data (pd.DataFrame): The data that we would like to sample.
            period (pd.Period): The period that we want to look at.
            freq (str): The sampling frequency that we would like to use.
            resample (str): An optional resampling function in case we need to
                resample the data we are given.
        """

        # First localize the data into the right timezone
        if self._check_tz_naive(data):
            data = data.tz_localize('UTC', copy=False)

        data = data.tz_convert(self._timezone, copy=False)

        # Create the final period index we will return
        index = self.divide_period(period, freq)

        # If the data sampling frequency does not match our desired
        # final sampling frequency, we need to resample it before
        # we can reindex it.
        if data.index.freq != index.freq:
            resampler = data.resample(freq)

            if isinstance(resample, str):
                data = getattr(resampler, resample)()
            else:
                data = resampler.apply(resample)

        return data.reindex(index)



""":mod:`iotile_analytics.core` - High level entry point for iotile_analytics.core package."""

from .group import AnalysisGroup
from .session import CloudSession
from .interaction import ProgressBar
from .environment import Environment

__all__ = ['AnalysisGroup', 'CloudSession', 'ProgressBar', 'Environment']



"""The AnalysisGroup is the main entry point for analyzing data.

It allows you to select data streams from IOTile.cloud and turn them
into Pandas time series for analysis.  You can create an AnalysisGroup
from a Project, a Device or a DataBlock.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from future.utils import viewitems, viewvalues

import pkg_resources
import pandas as pd
from iotile_cloud.api.connection import DOMAIN_NAME
from iotile_cloud.api.exceptions import RestHttpBaseException
from iotile_cloud.stream.data import StreamData
from typedargs.exceptions import ArgumentError
from .exceptions import CloudError
from .session import CloudSession
from .channels import IOTileCloudChannel


class AnalysisGroup(object):
    """A top level entry point for selecting and analyzing data.

    Currently AnalysisGroups must be created by passing the id
    or slug of a project in iotile.cloud and then are populated with
    all of the data streams and variables in that project.

    You should use one of the static methods FromDevice, FromArchive,
    FromProject or FromFile.

    If you use FromDevice, you should pass a slug for the device that you want
    to fetch streams from.

    If you use FromArchive you should pass a slug for the archive
    that you want to fetch streams from.

    If you use FromProject, you should pass the UUID of the project
    that you want to fetch streams from.

    Args:
        channel (AnalysisGroupChannel): The channel by which we can
            find and download streams.
    """

    def __init__(self, channel):
        self._channel = channel

        stream_list = channel.list_streams()
        self.source_info = channel.fetch_source_info()
        self.properties = channel.fetch_properties()
        self.streams = self._parse_stream_list(stream_list)
        self.stream_counts = channel.count_streams([x['slug'] for x in stream_list])

        var_type_slugs = set([x['var_type'] for x in viewvalues(self.streams) if x['var_type'] is not None])

        self.variable_types = channel.fetch_variable_types(var_type_slugs)
        self._stream_table = [(slug.lower(), self.get_stream_name(stream).lower()) for slug, stream in viewitems(self.streams)]

    def stream_empty(self, slug):
        """Check if a stream is empty.

        Empty streams have neither event data nor time series data.

        Args:
            slug (str): The slug of the stream tha we want to check to
                see if it empty

        Returns:
            bool: True if the stream is empty, otherwise false.
        """

        counts = self.stream_counts[slug]
        return counts['points'] == 0 and counts['events'] == 0

    def print_source_info(self):
        """Print a table with source object info

        The source object is the Project, Device or DataBlock this channel was created from

        Args:
            None
        """

        print("{:30s} {:s}".format("Name", "Value"))
        print("{:30s} {:s}".format("----", "----"))
        new_line = '\n' + ' ' * 31
        for key in self.source_info.keys():
            if len(key) > 37:
                key = key[:37] + '...'

            print('{0:30s} {1}'.format(key, str(self.source_info[key]).replace('\n', new_line)))

    def print_streams(self, include_empty=False):
        """Print a table of the streams in this AnalysisGroup.

        By default this function does not print streams that have no data
        but you can override that by passing include_empty=True.

        Args:
            include_empty (bool): Also show streams that have no data.
        """

        print("{:40s} {:s}".format("Name", "Slug"))
        print("{:40s} {:s}".format("----", "----"))

        for slug, stream in self.streams.items():
            if self.stream_empty(slug) and not include_empty:
                continue

            name = self.get_stream_name(stream)

            if len(name) > 37:
                name = name[:37] + '...'

            print('{:40s} {:s}'.format(name, slug))

    @classmethod
    def get_stream_name(cls, stream):
        """Extract a user-friendly name from a stream.

        Args:
            stream (dict): A dictionary of metadata about the stream.
        Returns:
            str: The name of the stream
        """

        name = stream.get('data_label', '')

        if name == '':
            name = stream.get('var_name') + "(from variable)"

        if name == '':
            name = "Unnamed"

        return name

    def find_stream(self, partial, include_empty=False):
        """Find a stream by a partial identifier.

        The identifier can either be a part of its slug or its name or the
        name of the variable that it comes from.  Whatever you pass must
        uniquely match exactly one stream in this analysis project or an
        exception will be thrown.

        Matching is done in a case insensitive fashion.

        By default, only streams that are not empty are searched.  You can
        override this by passing include_empty=True.  This behavior is the
        same as the print_streams() behavior, so searching for a stream based
        on what you find in print streams will work as expected.

        Args:
            partial (str): A partial slug or identifier to use to uniquely
                find one stream.
            include_empty (bool): Also include streams that have no data in
                the search.

        Returns:
            str: The full slug of the stream that we were looking for, suitable
                for passing to an IOTile.cloud api.

        Raises:
            ArgumentError: If the partial matches no streams or if it matches
                multiple streams.
        """

        if isinstance(partial, int):
            partial = "{:x}".format(partial)

        if not isinstance(partial, str):
            partial = str(partial)

        partial = partial.lower()

        found = [slug for slug, stream in self._stream_table if (partial in slug or partial in stream) and (include_empty or not self.stream_empty(slug))]
        if len(found) == 0:
            raise ArgumentError("No stream found matching given partial identifier", partial=partial)
        elif len(found) > 1:
            raise ArgumentError("Multiple matching streams found", partial=partial, slugs=found)

        return found[0]

    def fetch_stream(self, slug_or_name):
        """Fetch data from a stream by its slug or name.

        For example say you have the following stream in this analysis project:
        s--0000-0011--0000-0000-0000-0222--5001 (Temperature)

        You could fetch it by passing any of:
         - temp
         - 5001
         - Temperature
         - s--0000-0011--0000-0000-0000-0222--5001

        Args:
            slug_or_name (str): The stream that we want to fetch.  This
                can be a partial match to a full stream slug or name so long
                as it uniquely matches.  This is passed to find_stream so anything
                that find_stream accepts will be accepted here.

        Returns:
            StreamSeries: A pandas DataFrame subclass containing the data points as columns.
                The index of the dataframe is time in UTC.
        """

        slug = self.find_stream(slug_or_name)
        stream = self.streams[slug]
        raw = self._channel.fetch_datapoints(slug)
        raw.set_stream(self.streams[slug])

        vartype_slug = stream['var_type']

        if vartype_slug is not None:
            vartype = self.variable_types[stream['var_type']]
            raw.set_vartype(vartype)

        return raw

    def fetch_events(self, slug_or_name):
        """Fetch event metadata from a stream by its slug or name.

        This function will return a Pandas DataFrame with all of the
        events in the stream.  It will return only the extra summary
        data stored with the event itself however.  If you need access to
        any raw data uploaded with the event that was not part of the
        summary, you need to use fetch_event for each event to get
        the raw data.

        slug_or_name is passed to find_stream to convert it into a slug so
        partial stream names are accepted.

        Args:
            slug_or_name (str): The stream that we want to fetch.  This
                can be a partial match to a full stream slug or name so long
                as it uniquely matches.  This is passed to find_stream so anything
                that find_stream accepts will be accepted here.

        Returns:
            DataFrame: All of the extra_data associated with the event as a
                Pandas DataFrame.
        """

        slug = self.find_stream(slug_or_name)
        return self._channel.fetch_events(slug)

    def fetch_raw_events(self, slug_or_name, subkey=None, postprocess=None):
        """Fetch multiple raw events by numeric id.

        Args:
            slug_or_name (str): The stream that we want to fetch.  This
                can be a partial match to a full stream slug or name so long
                as it uniquely matches.  This is passed to find_stream so anything
                that find_stream accepts will be accepted here.
            subkey (str): Only include a single key of the event.  This is usefuly for complex
                events where you want to just focus on a portion of them.
            postprocess (callable): (Optional) function to call on each raw event before
                adding it to the dataframe.

        Returns:
            pd.DataFrame: The raw event object data fetched from iotile.cloud.
        """

        slug = self.find_stream(slug_or_name)

        raw_events = self._channel.fetch_raw_events(slug)

        if postprocess is None:
            return raw_events

        if subkey is not None:
            data = [postprocess(x[subkey]) for _i, x in raw_events.iterrows()]
        else:
            data = [postprocess(x) for _i, x in raw_events.iterrows()]

        return pd.DataFrame(data, index=raw_events.index)

    @classmethod
    def _parse_stream_list(cls, stream_list):
        return {stream['slug']: stream for stream in stream_list}

    @classmethod
    def FromDevice(cls, slug=None, external_id=None, domain=DOMAIN_NAME):
        """Create a new AnalysisGroup from a single device.

        The device can be found either by passing its slug or numeric
        id directly or by passing a device specific external_id.

        External ids are typically used to link a virtual IOTile device
        with a separate physical object.  If you have a POD or Arch Node
        then you don't need to worry about external ids.

        **You must pass either a slug or an external_id but not both.**

        Args:
            slug (str): the slug of the device that we want.  This can either
                 be a short slug with leading zeros omitted or long slug. It
                 should start with 'd--' since it is a device slug.
            external_id (str): The external id of the device that we want to
                search for. This external_id must be unique among all devices
                so that it matches a single device.
            domain (str): Optional iotile.cloud domain to connect to (defaults to
                https://iotile.cloud).
        """

        session = CloudSession(domain=domain)
        device = session.find_device(slug=slug, external_id=external_id)

        channel = IOTileCloudChannel(device['slug'], source_type="device", domain=domain)
        return AnalysisGroup(channel)

    @classmethod
    def FromArchive(cls, slug=None, domain=DOMAIN_NAME):
        """Create a new AnalysisGroup from a single archive.

        The archive is found by passing its archive slug:
        b--XXXX-XXXX-XXXX-XXXX

        Args:
            slug (str): the slug of the archive that we want.  This can either
                 be a short slug with leading zeros omitted or long slug. It
                 should start with 'b--' since it is an archive slug.
            domain (str): Optional iotile.cloud domain to connect to (defaults to
                https://iotile.cloud).
        """

        channel = IOTileCloudChannel(slug, source_type="datablock", domain=domain)
        return AnalysisGroup(channel)

    @classmethod
    def FromSaved(cls, identifier, format_name):
        """Load a saved AnalysisGroup.

        You must have previously called save on an AnalysisGroup
        and passed it a known format_name and identifier.  Passing
        that same identifier and format_name will reload that same
        AnalysisGroup without needing online access to IOTile.cloud.

        Args:
            identifier (str): The format specific identifier that we will
                use to name this saved AnalysisGroup.  The meaning of this
                parameter depends on the format but, for example, it could
                be the file name.
            format_name (str): An identifier for the format we wish to use to
                save our data.  You can find the list of available formats
                by calling list_formats().

        Returns:
            AnalysisGroup: The saved AnalysisGroup
        """

        loader = cls._find_load_format(format_name)

        channel = loader(identifier)
        return AnalysisGroup(channel)

    def save(self, identifier, format_name):
        """Save this AnalysisGroup.

        You can then load this analysis group again by calling
        AnalysisGroup.FromSaved(identifier, format)

        with the same identifier and format information.

        Args:
            identifier (str): The format specific identifier that we will
                use to name this saved AnalysisGroup.  The meaning of this
                parameter depends on the format but, for example, it could
                be the file name.
            format_name (str): An identifier for the format we wish to use to
                save our data.  You can find the list of available formats
                by calling list_formats().
        """

        saver_factory = self._find_save_format(format_name)
        with saver_factory(identifier) as saver:
            for slug, stream in viewitems(self.streams):
                data = None
                events = None
                raw_events = None
                if not self.stream_empty(slug):
                    data = self.fetch_stream(slug)
                    events = self.fetch_events(slug)
                    raw_events = self.fetch_raw_events(slug)

                saver.save_stream(slug, stream, data, events, raw_events)

            for slug, vartype in viewitems(self.variable_types):
                saver.save_vartype(slug, vartype)

            saver.save_source_info(self.source_info, self.properties)

    @classmethod
    def _find_save_format(cls, format_name):
        for entry in pkg_resources.iter_entry_points('iotile_analytics.save_format', format_name):
            return entry.load()

        raise ArgumentError("Could not find output format by name", available_formats=cls.list_formats())

    @classmethod
    def _find_load_format(cls, format_name):
        for entry in pkg_resources.iter_entry_points('iotile_analytics.load_format', format_name):
            return entry.load()

        raise ArgumentError("Could not find output format by name", available_formats=cls.list_formats())

    @classmethod
    def list_formats(cls):
        """List the installed formats in which we can save an AnalysisGroup.

        Returns:
            list(str): A list of the names that can be passed to save or FromSaved.
        """

        return [x.name for x in pkg_resources.iter_entry_points('iotile_analytics.save_format')]



"""Methods by which AnalysisGroup objects can find and download streams."""


class AnalysisGroupChannel(object):
    """A delegate object that can find and download streams.

    AnalysisGroupChannels must provide ways to locate streams,
    download time series data and event data for them as well
    as a few other routines.  AnalysisGroups delegate requests
    for streams to their internal AnalysisGroupChannel so that
    you can access IOTile data that is either stored locally
    or remotely.
    """

    def list_streams(self):
        """Return a list of all streams.

        This is equivalent to the IOTile.cloud API method
        /api/v1/stream/

        Returns:
            list(dict): A list of dictionaries, one for each
                stream that should be part of this analysis group.
        """

        raise NotImplementedError()

    def count_streams(self, slugs):
        """Count the number of events and data points in a stream.

        Args:
            slugs (list(str)): The slugs of the stream that we should count.

        Returns:
            dict(<slug>: {'points': int, 'events': int}): A dict mapping dicts of 2
                integers with the count of the number of events and the number of
                data points in this stream.
        """

        raise NotImplementedError()

    def fetch_variable_types(self, slugs):
        """Fetch variable type information for a list of variable slugs.

        Args:
            slugs (list(str)): The slugs of the variable types that we should fetch.

        Returns:
            dict(<slug>: dict): A dict mapping variable slugs to variable type definitions
        """

        raise NotImplementedError()

    def fetch_events(self, slug):
        """Fetch all events for a given stream.

        These are the event metadata dictionaries, not the raw
        event data that may be stored along with the metadata.

        Args:
            slug (str): The slug of the stream that we should fetch
                events for.

        Returns:
            pd.DataFrame: All of the events.
        """

        raise NotImplementedError()

    def fetch_raw_events(self, slug):
        """Fetch all raw event data for this stream.

        These are the raw json dictionaries that are stored for
        each event.

        Args:
            slug (str): The slug of the stream that we should fetch
                raw events for.

        Returns:
            pd.DataFrame: All of the raw events.
        """

        raise NotImplementedError()

    def fetch_source_info(self):
        """Fetch the record associated to the channel object (project, device or datablock)

        This is the object dictionary for the project, device or datablock this channel is based on

        Returns:
            dict(<name>: <value>): A dict mapping object attribute names and values.
        """

        raise NotImplementedError()

    def fetch_properties(self):
        """Fetch all properties for a given object (project, device or datablock).

        Returns:
            dict: A dict of property names and values.
        """

        raise NotImplementedError()

    def fetch_datapoints(self, slug, direct=False):
        """Fetch all data points for this stream.

        These are time, value data pairs stored in the stream.

        Args:
            slug (str): The slug of the stream that we should fetch
                raw events for.
            direct (bool): Access the data directly without needing a
                stream object to perform unit conversion.

        Returns:
            StreamSeries: A data fame with internal value as floating
                point data.
        """

        raise NotImplementedError()



"""A wrapper around the iotile.cloud API to fetch data diretcly."""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from future.utils import viewitems

import pandas as pd
from iotile_cloud.stream.data import StreamData
from iotile_cloud.api.exceptions import RestHttpBaseException
from typedargs.exceptions import ArgumentError
from .channel import AnalysisGroupChannel
from ..session import CloudSession
from ..interaction import ProgressBar
from ..stream_series import StreamSeries
from ..exceptions import CloudError


class IOTileCloudChannel(AnalysisGroupChannel):
    """An AnalysisGroupChannel that fetches data from IOTile.cloud

    Args:
        domain (str): The domain of the iotile.cloud instance that
            you want to fetch data from.
        cloud_id (str): The slug or id of the object to be enumerated to
            create this cloud channel.
        source_type (str): The type of object that we are enumerating.
            This must be one of project, device or archive.
    """

    def __init__(self, cloud_id, source_type, domain):
        super(IOTileCloudChannel, self).__init__()

        stream_finders = {
            'project': self._find_project_streams,
            'device': self._find_device_streams,
            'datablock': self._find_archive_streams
        }

        stream_finder = stream_finders.get(source_type, None)
        if stream_finder is None:
            raise ArgumentError("Invalid source type", source_type=source_type, supported_sources=stream_finders.keys(), suggestion="Try using one of the convenience functions for creating an AnalysisGroup like FromDevice or FromProject")

        self._session = CloudSession(domain=domain)
        self._api = self._session.get_api()
        self._cloud_id = cloud_id
        self._source_type = source_type
        self._stream_finder = stream_finder

    def list_streams(self):
        """Return a list of all streams.

        This is equivalent to the IOTile.cloud API method
        /api/v1/stream/

        Returns:
            list(dict): A list of dictionaries, one for each
                stream that should be part of this analysis group.
        """

        return self._stream_finder(self._cloud_id)

    def fetch_source_info(self):
        """Fetch the record associated to the channel object (project, device or datablock)

        This is the object dictionary for the project, device or datablock
        this channel is based on

        Args:
            with_properties: If True, will also fetch object properties and
                add them as dictionary entries

        Returns:
            dict: The raw source object that our analytics group was generated from.
        """

        resource = getattr(self._api, self._source_type)
        try:
            data = resource(self._cloud_id).get()
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)

        return data

    def fetch_properties(self):
        """Fetch all properties for a given object (project, device or datablock).

        Returns:
            dict: A dict of property names and values.
        """

        data = {}

        try:
            property_data = self._api.property.get(target=self._cloud_id)
            if 'count' in property_data and property_data['count']:
                for item in property_data['results']:
                    data[item['name']] = item['value']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc,
                             response=exc.response.status_code)

        return data

    def count_streams(self, slugs):
        """Count the number of events and data points in a stream.

        Args:
            slug (list(str)): The slugs of the stream that we should count.

        Returns:
            dict(<slug>: {'points': int, 'events': int}): A dict mapping dicts of 2
                integers with the count of the number of events and the number of
                data points in this stream.
        """

        resources = [self._api.stream(x).data for x in slugs]
        ts_results = self._session.fetch_multiple(resources, message='Counting Events in Streams', page_size=1)

        event_resources = [self._api.event for x in slugs]
        event_args = [{"filter": x} for x in slugs]

        event_results = self._session.fetch_multiple(event_resources, event_args, message='Counting Data in Streams', page_size=1)

        return {slug: {'points': ts['count'], 'events': event['count']} for slug, ts, event in zip(slugs, ts_results, event_results)}

    def fetch_variable_types(self, slugs):
        """Fetch variable type information for a list of variable slugs.

        Args:
            slugs (list(str)): The slugs of the variable types that we should fetch.

        Returns:
            dict(<slug>: dict): A dict mapping variable slugs to variable type definitions
        """

        resources = [self._api.vartype(x) for x in slugs]
        variables = self._session.fetch_multiple(resources, message='Fetching Variable Types')

        return {x['slug']: x for x in variables}

    def fetch_events(self, slug):
        """Fetch all events for a given stream.

        These are the event metadata dictionaries, not the raw
        event data that may be stored along with the metadata.

        Args:
            slug (str): The slug of the stream that we should fetch
                events for.

        Returns:
            pd.DataFrame: A data frame with all of the events in this
                stream.
        """

        try:
            resource = self._api.event
            data = self._session.fetch_all(resource, page_size=1000, message="Downloading Events", filter=slug)

            dt_index = pd.to_datetime([x['timestamp'] for x in data])
            extra_data = [x['extra_data'] for x in data]

            for i, event in enumerate(data):
                extra_data[i]['event_id'] = event['id']
                extra_data[i]['has_raw_data'] = event.get('has_raw_data', False)

            return pd.DataFrame(extra_data, index=dt_index)
        except RestHttpBaseException as exc:
            raise CloudError("Error fetching events from stream", exception=exc, response=exc.response.status_code)

    def fetch_raw_events(self, slug):
        """Fetch all raw event data for this stream.

        These are the raw json dictionaries that may be stored for
        each event.

        Args:
            slug (str): The slug of the stream that we should fetch
                raw events for.

        Returns:
            pd.DataFrame: All of the raw events.  This will be empty if
                there are no raw events.
        """

        events = self.fetch_events(slug)
        if len(events) == 0:
            return pd.DataFrame()

        event_ids = events['event_id'].values
        # event.has_raw_data is true if there is an associated raw JSON file stored on S3 for this event
        event_has_data = events['has_raw_data'].values
        indexes = events.index
        resources = []
        new_index = []
        for index in range(len(event_ids)):
            if event_has_data[index]:
                resources.append(self._api.event(event_ids[index]).data)
                new_index.append(indexes[index])

        data = self._session.fetch_multiple(resources, message="Downloading Raw Event Data")
        # new_index includes all indexes with actual data
        new_index = pd.to_datetime(new_index)
        return pd.DataFrame(data, index=new_index)

    def fetch_datapoints(self, slug):
        """Fetch all data points for this stream.

        These are time, value data pairs stored in the stream.

        Args:
            slug (str): The slug of the stream that we should fetch
                raw events for.

        Returns:
            StreamSeries: A data fame with internal value as floating
                point data.
        """

        with ProgressBar(1, "Fetching %s" % slug, leave=False) as prog:
            raw_data = self._api.df.get(filter=slug, format='csv')

            str_data = raw_data.decode('utf-8')
            rows = str_data.splitlines()
            data = [x.split(',') for x in rows]
            data = data[1:]  # There is a single header line
            dt_index = pd.to_datetime([x[0] for x in data])
            prog.update(1)

        return StreamSeries([float(x[1]) for x in data], index=dt_index)

    def _find_device_streams(self, device_slug):
        """Find all streams for a device by its slug."""

        try:
            results = self._api.stream.get(device=device_slug)
            return results['results']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)

    def _find_archive_streams(self, archive_slug):
        """Find all streams for an archive by its slug."""

        try:
            results = self._api.stream.get(block=archive_slug, all=1)
            return results['results']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)

    def _find_project_streams(self, project_id):
        """Find all streams in a project by its uuid."""

        try:
            results = self._api.stream.get(project=project_id)
            return results['results']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)



"""Custom exceptions for iotile analytics."""

from typedargs.exceptions import KeyValueException


class CloudError(KeyValueException):
    """A generic error interacting with IOTile.cloud."""

    pass


class AuthenticationError(CloudError):
    """An error logging into iotile.cloud."""

    pass


class CertificateVerificationError(CloudError):
    """The iotile.cloud instance you are talking to does not have a trusted certificate.

    If you really want to talk to this cloud server, you can pass verify=False when you
    create a CloudSession in order to turn off certificate verification.
    """

    pass


class MissingPackageError(KeyValueException):
    """The operation you requested requires a packgae to be installed.

    The suggestion keyword argument should be the name of the package.
    """

    pass


class UsageError(KeyValueException):
    """A method was called in an unacceptable way.

    This exception is thrown when an API function is misused.
    """

    pass
